! CGV/solution/parts/cgvp00.F90

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH
! >>>>>>>>>>>>>>>>>>>>>> begin of task 00 <<<<<<<<<<<<<<<<<<<<<<

PROGRAM cgv_fortran

#ifndef serial
  INCLUDE 'mpif.h'
#endif

! ****************** declarations and global data ***************

! --- options: ---
INTEGER :: m,n         ! size of the physical-heat matrix without borders  
                     ! m=vertical, n=horizontal 
INTEGER :: iter_max    ! maximal number of iterations in CG
REAL :: epsilon        ! Limit for sqr( || residual vector || ) 
INTEGER :: print_level ! print_level = 1: Only result-error and partial result matrix 
                                 ! = 2: and residual norm after each iteration 
                                 ! = 3: and result physical-heat matrix 
                                 ! = 4: and all vector and halo data in first iteration 
                                 ! = 5: and all vector and halo data in _all_ iterations
INTEGER :: decomp_dims      ! dimensins of the domain decomposition 
INTEGER :: m_procs, n_procs ! number of processes in both directions of the 2-dim domain decomposition 
INTEGER :: mprt, nprt       ! number of rows and columns when printing the partial result matrix 

! global structures and data: 

#ifndef serial
  INTEGER :: ierror, comm_cart
  DOUBLE PRECISION :: commtime_start=0, commtime_end=0, commtime_sum=0
#endif
INTEGER :: numprocs, my_rank
REAL :: flops

! size of the physical area this process is about to calculate
! start_i   : start row
! end1_i    : end row + 1
! start_j   : start column
! end1_j    : end column + 1
! rowsize   : number of data entries in a row of the local chunk
!             (end1_j - start_j)
! colsize   : number of data entries in a column of the local chunk
!             (end1_i - start_i)
! chunksize : number of data entries in the local chunk PLUS halo data
!             ((rowsize + 2) * (colsize + 2))
INTEGER :: start_i, end1_i, start_j, end1_j, rowsize, colsize, chunksize

! SIDX (Serial InDeX): computes the serial index of a two-dimensional array 
!   input: i = 0...(colsize + 1) and j = 0...(rowsize + 1) 
!   output: SIDX in 0..((colsize + 2) * (rowsize + 2) - 1)
#define SIDX(i, j) ((rowsize + 2) * (i) + (j))

! --- data structures ---

! halo structure
TYPE halo_struct
    REAL, DIMENSION(:), POINTER :: north
    INTEGER                     :: north_rank
    INTEGER                     :: north_size
    REAL, DIMENSION(:), POINTER :: east
    INTEGER                     :: east_rank
    INTEGER                     :: east_size
    REAL, DIMENSION(:), POINTER :: south
    INTEGER                     :: south_rank
    INTEGER                     :: south_size
    REAL, DIMENSION(:), POINTER :: west
    INTEGER                     :: west_rank
    INTEGER                     :: west_size
END TYPE halo_struct
TYPE(halo_struct) :: halo

! needed global vectors
REAL, DIMENSION(:), ALLOCATABLE :: x, b
! needed send and receive scratch buffers
REAL, DIMENSION(:), ALLOCATABLE :: recv_scratch_buff_east, recv_scratch_buff_west, &
                                 & recv_scratch_buff_north, recv_scratch_buff_south

! >>>>>>>>>>>>>>>>>>>>>> -end- of task 00 <<<<<<<<<<<<<<<<<<<<<<
