! CGV/solution/parts/cgvp03.F90

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH
!>>>>>>>>>>>>>>>>>>>>>> begin of task 03 <<<<<<<<<<<<<<<<<<<<<<

!********************** matrix-vector-multiply ****************************

!------------------------------
! mult_A_vec
!      IN  REAL(0:) v1 : vector to multiply A with
!      OUT REAL(0:) v2 : output vector
!
! TASK: Calculates matrix vector product v2 = A * v1
!
! USED GLOBAL DATA: 
!   IN     INTEGER rowsize: number of data entries in a row of the local chunk
!   IN     INTEGER colsize: number of data entries in a column of the local chunk
!   IN/OUT REAL    flops  : amount of FLOPS
!
! USED FUNCTIONS:
!   --
SUBROUTINE mult_A_vec(v1, v2)
    REAL, DIMENSION(0:), INTENT(IN)  :: v1
    REAL, DIMENSION(0:), INTENT(OUT) :: v2
    INTEGER :: i, j

#ifdef FASTMEMXS
    INTEGER :: k, calcend
#endif

    !**************************************************************************
    !* Matrix-Vector mult. without distinction of border/halo neighbored data *
    !**************************************************************************

#ifdef FASTMEMXS
    ! multiplication only for the inner values
    calcend = (colsize + 1) - 2 - 1
    IF (MODULO(calcend, 2) == 0) THEN
        calcend = calcend - 1
    END IF
    DO i = 1, calcend, 2
        DO j = 1, (rowsize + 1) - 1
            v2(SIDX(i, j)) = -1 * v1(SIDX((i - 1), j)) &
                          & - 1 * v1(SIDX(i, (j - 1))) &  
                          & + 4 * v1(SIDX(i, j)) &
                          & - 1 * v1(SIDX(i, (j + 1))) &
                          & - 1 * v1(SIDX((i + 1), j))
            v2(SIDX(i + 1, j)) = -1 * v1(SIDX((i - 1 + 1), j)) &
                          & - 1 * v1(SIDX((i + 1), (j - 1))) &  
                          & + 4 * v1(SIDX((i + 1), j)) &
                          & - 1 * v1(SIDX((i + 1), (j + 1))) &
                          & - 1 * v1(SIDX((i + 1 + 1), j))
        END DO
    END DO
 
    ! calculate the remaining rows
    DO k = calcend + 2, (colsize + 1) - 1
        DO j = 1, (rowsize + 1) - 1
            v2(SIDX(k, j)) = -1 * v1(SIDX((k - 1), j)) &
                          & - 1 * v1(SIDX(k, (j - 1))) &
                          & + 4 * v1(SIDX(k, j)) &
                          & - 1 * v1(SIDX(k, (j + 1))) &
                          & - 1 * v1(SIDX((k + 1), j))
        END DO
    END DO
#endif
#ifndef FASTMEMXS
    ! multiplication only for the inner values
    DO i = 1, (colsize + 1) - 1
        DO j = 1, (rowsize + 1) - 1
            v2(SIDX(i, j)) = -1 * v1(SIDX((i - 1), j)) &
                          & - 1 * v1(SIDX(i, (j - 1))) &  
                          & + 4 * v1(SIDX(i, j)) &
                          & - 1 * v1(SIDX(i, (j + 1))) &
                          & - 1 * v1(SIDX((i + 1), j))
        END DO
    END DO
#endif

    ! for the MFLOPS calculation
    flops = flops + rowsize * colsize * 9
END SUBROUTINE mult_A_vec

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 03 <<<<<<<<<<<<<<<<<<<<<<
