! CGV/solution/parts/cgvp05.F90

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH
!>>>>>>>>>>>>>>>>>>>>>> begin of task 05 <<<<<<<<<<<<<<<<<<<<<<
!************* boundary and exact solution init ****************

!------------------------------
! Init_Boundary_Vec
!  OUT REAL(0:) b  : the boundary vector 
!
! TASK: Store all boundary values in the vector memory
!       (that is already allocated).
!       In this example, the boundaries are given by:
!         u(i,j) = h * (i+1)  with  h = 1.0/(m+1) for i=-1,m and j=-1,n
!
! GLOBAL DATA:
!   IN INTEGER start_i : start row of the physical area this process is to calculate
!   IN INTEGER end1_i  : end row + 1 of the physical area this process is to calculate
!   IN INTEGER start_j : start column of the physical area this process is to calculate
!   IN INTEGER end1_j  : end column + 1 of the physical area this process is to calculate
!   IN INTEGER m       : vertical size of the physical problem
!   IN INTEGER n       : horizontal size of the physical problem
!
! USED FUNCTIONS:
!     --
SUBROUTINE Init_Boundary_Vec(b)
    REAL, DIMENSION(0:), INTENT(OUT) :: b
    INTEGER                          :: i, j
    REAL                             :: v, h

    h = 1.0/(m+1)
    ! boundary vector data is all zero at this point
    DO i = start_i, end1_i - 1
        DO j = start_j, end1_j - 1
            v = 0.0
            ! next line not neccessary, because boundary value is zero
            ! if (i==0)   v = v + u(-1,j) = 0
            ! next line simplified to add 1.0 to vector, because of the boundary value of 1.0
            ! if (i==m-1) v = v + u(m, j): h * (m+1) = 1
            IF (i==m-1) THEN
                v = 1.0
            END IF
            ! u(i,-1):  h * (i+1)
            IF (j==0) THEN
                v = v + h * (i + 1)
            END IF
            ! u(i, n):  h * (i+1)
            IF (j==n-1) THEN
                v = v + h * (i + 1)
            END IF
            b(SIDX((i - start_i + 1), (j - start_j + 1))) = v
        END DO
    END DO
END SUBROUTINE Init_Boundary_Vec

!------------------------------
! Init_Exact_Solution_Vec
!   OUT REAL(0:) u  : the local chunk of the exact solution vector
!
! TASK: Store the exact solution for the given boundary in the vector u.
!       In this example, the inner values are given by:
!       u(i,j) = h * (i+1)  with  h = 1.0/(m+1) for i=0..m-1 and j=0..n-1
!
! GLOBAL DATA:
!   IN INTEGER start_i : start row of the physical area this process is to calculate
!   IN INTEGER end1_i  : end row + 1 of the physical area this process is to calculate
!   IN INTEGER start_j : start column of the physical area this process is to calculate
!   IN INTEGER end1_j  : end column + 1 of the physical area this process is to calculate
!   IN INTEGER m       : vertical size of the physical problem
!
! USED FUNCTIONS:
!   --
SUBROUTINE Init_Exact_Solution_vec(u)
    REAL, DIMENSION(0:), INTENT(OUT) :: u
    INTEGER                          :: i, j
    REAL                             :: h

    h = 1.0 / (m + 1)
    DO i = start_i, end1_i - 1
        DO j = start_j, end1_j - 1
            u(SIDX((i - start_i + 1), (j - start_j + 1))) = h * (i + 1)
        END DO
    END DO
END SUBROUTINE Init_Exact_Solution_vec
!>>>>>>>>>>>>>>>>>>>>>> -end- of task 05 <<<<<<<<<<<<<<<<<<<<<<
