! CGV/solution/parts/cgvp06.F90

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH
!>>>>>>>>>>>>>>>>>>>>>> begin of task 06 <<<<<<<<<<<<<<<<<<<<<<

!******************* print routines ********************************
SUBROUTINE appl_print_vec(name, v, mprint, nprint)
    CHARACTER, INTENT(IN)              :: name
    REAL, DIMENSION(0:,0:), INTENT(IN) :: v
    INTEGER, INTENT(IN)                :: mprint, nprint
    INTEGER                            :: i,j
    IF (my_rank == 0) THEN
        WRITE(*,'(A, A)', ADVANCE='NO') name, ':       j='
        DO j=0, n - 1 
            IF ( (n <= nprint) .OR. ( (n > nprint) .AND. ((nprint-1)*(j+n-2)/(n-1) /= (nprint-1)*(j+n-1)/(n-1)))) THEN
                WRITE(*,'(I6)', ADVANCE='NO') j
            END IF
        END DO
        WRITE(*,*)
        DO i=0, m - 1
            IF ( (m <= mprint) .OR. ( (m > mprint) .AND. ((mprint-1)*(i+m-2)/(m-1) /= (mprint-1)*(i+m-1)/(m-1))  ) ) THEN
                WRITE(*,'(A, A, I3, A)', ADVANCE='NO') name, ': i=', i, '   '
                DO j=0, n - 1
                   IF ( (n <= nprint) .OR. ((n > nprint) .AND. ((nprint-1)*(j+n-2)/(n-1) /= (nprint-1)*(j+n-1)/(n-1)))) THEN
                       WRITE(*, '(A, F5.3)', ADVANCE='NO') ' ', v(i, j)
                   END IF
                END DO
                WRITE(*,*)
            END IF
        END DO
    END IF
END SUBROUTINE appl_print_vec

SUBROUTINE recv_token()
#ifndef serial
    INTEGER :: dummy, status(MPI_STATUS_SIZE)
    IF (my_rank > 0) THEN
        CALL MPI_RECV(dummy, 0, MPI_INTEGER, my_rank-1, 333, comm_cart, status, ierror)
    END IF
#endif
END SUBROUTINE recv_token

SUBROUTINE flush_and_send_token()
#ifndef serial
    INTEGER    :: dummy, status(MPI_STATUS_SIZE)
    CALL MPI_SEND(dummy, 0, MPI_INTEGER, MODULO((my_rank+1), numprocs), 333, comm_cart, ierror)
    ! Process 0 must wait until the last process (numprocs-1) has finished its output
    IF (my_rank == 0) THEN
        CALL MPI_RECV(dummy, 0, MPI_INTEGER, numprocs-1, 333, comm_cart, status, ierror)
    END IF
#endif
END SUBROUTINE flush_and_send_token

SUBROUTINE print_halo()
#ifndef serial
    INTEGER :: i

    CALL recv_token()

    WRITE(*,*)
    WRITE(*,*) 'Halo of rank ', my_rank
    WRITE(*,*) '----------------'
    IF (halo%north_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'North: NO halo'
    ELSE
        WRITE(*,*) 'North:'
        WRITE(*,*) '------'
        WRITE(*,*) 'values: ', halo%north_size, ', neighbor rank: ', halo%north_rank
        DO i = 1, halo%north_size
            WRITE(*,*) 'halo[', i, ']: ', halo%north(i)
        END DO
    END IF
    IF (halo%east_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'East: NO halo'
    ELSE
        WRITE(*,*) 'East:'
        WRITE(*,*) '------'
        WRITE(*,*) 'values: ', halo%east_size, ', neighbor rank: ', halo%east_rank
        DO i = 1, halo%east_size
            WRITE(*,*) 'halo[', i, ']: ', halo%east(i)
        END DO
    END IF
    IF (halo%south_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'South: NO halo'
    ELSE
        WRITE(*,*) 'South:'
        WRITE(*,*) '------'
        WRITE(*,*) 'values: ', halo%south_size, ', neighbor rank: ', halo%south_rank
        DO i = 1, halo%south_size
            WRITE(*,*) 'halo[', i, ']: ', halo%south(i)
        END DO
    END IF
    IF (halo%west_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'West: NO halo'
    ELSE
        WRITE(*,*) 'West:'
        WRITE(*,*) '------'
        WRITE(*,*) 'values: ', halo%west_size, ', neighbor rank: ', halo%west_rank
        DO i = 1, halo%west_size
            WRITE(*,*) 'halo[', i, ']: ', halo%west(i)
        END DO
    END IF
    WRITE(*,*) ''
    CALL flush_and_send_token()
#endif
END SUBROUTINE print_halo

SUBROUTINE print_halo_struct()
#ifndef serial
    CALL recv_token()

    WRITE(*,*) ''
    WRITE(*,*) 'Halo of rank ', my_rank
    WRITE(*,*) '----------------'
    IF (halo%north_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'North: NO halo'
    ELSE
        WRITE(*,*) 'North: values: ', halo%north_size ,', neighbor rank: ', halo%north_rank
    END IF
    IF (halo%east_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'East: NO halo'
    ELSE
        WRITE(*,*) 'East: values: ', halo%east_size ,', neighbor rank: ', halo%east_rank
    END IF
    IF (halo%south_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'South: NO halo'
    ELSE
        WRITE(*,*) 'South: values: ', halo%south_size ,', neighbor rank: ', halo%south_rank
    END IF
    IF (halo%west_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'West: NO halo'
    ELSE
        WRITE(*,*) 'West: values: ', halo%west_size ,', neighbor rank: ', halo%west_rank
    END IF
    WRITE(*,*) ''

    CALL flush_and_send_token()
#endif
END SUBROUTINE print_halo_struct

SUBROUTINE print_vec2d(name, v2, mprint, nprint)
    CHARACTER, INTENT(IN)           :: name
    REAL, DIMENSION(0:), INTENT(IN) :: v2
    INTEGER, INTENT(IN)             :: mprint, nprint
    INTEGER :: i,j
    REAL    :: whole_v2(0:m-1, 0:n-1)
#ifndef serial
    INTEGER    :: rank
    REAL       :: buff(0:(n + 2) * (m + 2) - 1)
    INTEGER    :: extent(0:3), new_offset
    INTEGER    :: status(MPI_STATUS_SIZE), status2(MPI_STATUS_SIZE)

    IF (my_rank > 0) THEN
        ! send the start_i, end1-i, start_j, end1_j
        extent(0) = start_i
        extent(1) = end1_i
        extent(2) = start_j
        extent(3) = end1_j
        CALL MPI_SEND(extent, 4, MPI_INTEGER, 0, 666, comm_cart, ierror)

        ! send the vector data to rank 0
        CALL MPI_SEND(v2, chunksize, MPI_REAL, 0, 333, comm_cart, ierror)
    ELSE
        ! copy the rank 0 chunk into the whole physical vector
        DO i = start_i, end1_i - 1
            DO j = start_j, end1_j - 1
                whole_v2(i, j) = v2(SIDX((i - start_i + 1), (j - start_j + 1)))
            END DO
        END DO
            
        ! receive the halo data and data extent from all other ranks into the buffer
        ! and copy the buffer to where it belongs into the whole physical vector
        DO rank = 1, numprocs - 1
            CALL MPI_RECV(extent, 4, MPI_INTEGER, rank, 666, comm_cart, status, ierror)
            CALL MPI_RECV(buff, ((extent(1) - extent(0) + 2) * (extent(3) - extent(2) + 2)), MPI_REAL, &
                        & rank, 333, comm_cart, status2, ierror)
    
            new_offset = extent(3) - extent(2) + 2
            
            DO i = extent(0), extent(1) - 1
                DO j = extent(2), extent(3) - 1
                    whole_v2(i, j) = buff((i - extent(0) + 1) * new_offset + (j - extent(2) + 1))
                END DO
            END DO
        END DO
    END IF
    
    CALL appl_print_vec(name, whole_v2, mprint, nprint)
#endif
#ifdef serial
    ! copy the data without halo into the whole physical vector
    DO i = start_i, end1_i - 1
        DO j = start_j, end1_j - 1
            whole_v2(i, j) = v2(SIDX((i - start_i + 1), (j - start_j + 1)))
        END DO
    END DO
    CALL appl_print_vec(name, whole_v2, mprint, nprint)
#endif
END SUBROUTINE print_vec2d

SUBROUTINE print_vec(name, v1)
    CHARACTER, INTENT(IN)           :: name
    REAL, DIMENSION(0:), INTENT(IN) :: v1
    INTEGER                         :: i, j
    CALL recv_token()
    DO i = 1, (colsize + 1) - 1
        DO j = 1, (rowsize + 1) - 1
            WRITE(*,'(A, A, I2, A, F6.3)') name, '[', ((i * rowsize) + j), '] = ', v1(SIDX(i, j))
        END DO
    END DO
    CALL flush_and_send_token()
END SUBROUTINE print_vec

SUBROUTINE print_decomp()
    CALL recv_token()
    WRITE(*, '(A,I3, A, I0, A, I0, A, I0, A, I0)') 'Rank ', my_rank, ': start_i, end1_i, start_j, end1_j: ', &
             & start_i, ', ', end1_i, ', ', start_j, ', ', end1_j
    CALL flush_and_send_token()
END SUBROUTINE print_decomp

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 06 <<<<<<<<<<<<<<<<<<<<<<
